--
-- LuaYacc -- Yet another yacc clone
-- a LARL Parser Generator for Lua
--
-- Copyright (c) 2006 Luis E. Garcia Ontanon <luis.ontanon@gmail.com>
--
-- This library is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2.1
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
--

-- The table generation algorithms have been taken from :
-- Aho, Sethi, Ullman. Compilers Principles, Techniques and Tools (ISBN 0-201-10194-7)
-- Comments in the code with reference to page numbers or sections refer to it.

-- TODO
  -- LARL
  -- integer tokens (are they worth?)
  -- option to generate "a-la" yacc parsers
     -- used like parser(next_token_function)
     -- instead of p = parser(); p(tok,val); p(tok2,val2); ...

-- SymStr
-- a symbol sequence
-- { s1, s2, s3, ...}
	mtSymStr = {}
	mtSymStr.__index = mtSymStr

-- Lr0 item
-- { rule, dotpos }
	mtLr0 = {}
	mtLr0.__index = mtLr0

-- Rule
-- {sym,symstr,orig_idx,prec}
	mtRule = {}
	mtRule.__index = mtRule

-- Grammar
-- { rules = {r1, r2, r3, ...}, term = {["A"] = 1, ["B"] = 1, ...}, nterm = {["a"] = 1, ["b"] = 1, ...}, ... }
	mtGrammar = {}
	mtGrammar.__index = mtGrammar

-- Lr1
-- {rule, dotpos, term}
	mtLr1 = {}
	mtLr1.__index = mtLr1

-- Set
-- a set containing unique constant items, for which key = tostring(value)
-- sets containing the elements with identical tostring(i) will be "__eq"
	mtSet = {}


-- outputs a string with lua code for the given object
function tolua(v,i)
	local t = type(v)
    local mt = getmetatable(v)
	local s = ""
	
	if i == nil then
		i = ""
	end

	if  t == "string" then
		s = s .. '"' .. string.gsub(v,'"','\\"') .. '"'
	elseif mt ~= nil and type(mt.__tolua) == "function" then
		s = s .. mt.__tolua(v)
	elseif t == "table" then
		s = s .. '\n' .. i .. '{\n'
		i = i .. "\t"
		for k,val in pairs(v) do
			s = s .. i .. "\t".. "[".. tolua(k).. "]" .. "=" .. tolua(val, i) .. ",\n"
		end
		s = s .. i .. "}"
	else
		s = s .. " " .. tostring(v) .." "
	end
	
	return s
end


function mtSet.__index(set)
	return set
end

function mtSet.__newindex(set)
	return function(s,k,v)
		rawset(set,k,v)
	end
end


function mtSet.insert(set)
-- tries to insert an item into this keeping the set sorted, allows for remove,
	local h = rawget(set,"h")
	local t = rawget(set,"t")

	return function(v)
		local k = tostring(v)
		local s = h[k]
		local pos = 0

		if s ~= nil then
			return false
		end

		h[k] = v

		for i,j in pairs(t) do
			if tostring(j) < k then
				break
			end
			pos = pos + 1
		end
		pos = pos + 1

		table.insert(t,pos,v)
		return true
	end
end

function mtSet.append(set)
-- creates a set in which items are kept in the order they were appeneded
-- once used the generated sets won't necessarilly be "__eq"
	local h = rawget(set,"h")
	local t = rawget(set,"t")

	return function(i)
		local k = tostring(i)
		local s = h[k]

		if s ~= nil then
			return false
		end

		h[k] = i

		table.insert(t,i)
		return true
	end
end

function mtSet.remove(set)
-- removes an object from the set
	local h = rawget(set,"h")
	local t = rawget(set,"t")

	return function(v)
		local k = tostring(v)

		h[k] = nil

		for i,j in pairs(t) do
			if tostring(v) == tostring(j) then
				table.remove(t,i)
				return
			end
		end

		return v
	end
end

function mtSet.has_changed(set)
-- returns true if the set has changed since the last time it was called
	local last_time_items = 0
	local t = rawget(set,"t")
	return function()
		local has_changed = false
		local current_items = table.getn(t)

		if current_items ~= last_time_items then
			has_changed =  true
		end

		last_time_items = current_items

		return has_changed
	end
end

function mtSet.insert_set(set)
-- inserts the elements of another set into this
	local insert = mtSet.insert(set)
	return function(other_set)
		for i in other_set() do
			insert(i)
		end
		return set
	end
end

function mtSet.len(set)
-- returns the number of elements in the set
	local t = rawget(set,"t")
	return function(x)
		return table.getn(t)
	end
end

function mtSet.__sub(set)
-- returns a copy of the lhs set where elements in lhs have been removed
	return function(l,r)
		local x = l.copy(l)
		local remove = x.remove

		for i in r() do
			remove(i)
		end
		
		return x
	end
end


function mtSet.__mul(set)
-- intersection of two sets
	local h = rawget(set,"h")

	return function(l,r)
		local x = Set()
		local insert = x.insert


		for i in r() do
			for z,y in pairs(h) do
				if tostring(i) == tostring(z) then
					insert(i)
				end
			end
		end
		
		return x
	end

end

function mtSet.copy(set)
-- returns a copy of this set (w/ references to the elements of the original set)
	local t = rawget(set,"t")
	return function()
		return Set(t)
	end
end

function mtSet.contains(set)
-- returns true if the set contains an item with identical tostring() of the given value
	local t = rawget(set,"t")
	return function(v)
		local k = tostring(v)

		for i,j in pairs(t) do
			if tostring(j) == k then return true end
		end
	end
end

function mtSet.__add(set)
-- returns a set containing the elemets of both sets
	local t = rawget(set,"t")

	return function(l,r)
		local x = set.copy()
		local insert = x.insert

		for i,v in pairs(t) do
			insert(v)
		end

		for i,v in pairs(rawget(r,"t")) do
			insert(v)
		end
		
		return x
	end

end


function mtSet.__call(set)
-- returns an iterator for the set, the second arg is a condition:
--   - a string will be used as pattern, only elements matching will be returned
--   - a function will be called with the elemnty as argumets and the element will be returned if it returns true
	local t = rawget(set,"t")

	return function(x,cond) 
		local i = 0
		local n = table.getn(t)
		local ct = type(cond)
		
		if ct == "function" then
			i = 1
			return function()
				if i > n then return end
				repeat
					if cond(t[i]) then
						i = i + 1
						return t[i-1]
					end
					i = i + 1
				until i > n
			end
		elseif ct == "string" then
			i = 1
			return function()
				if i > n then return end
				repeat
					local v = t[i]
					if string.match(tostring(v),cond) ~= nil then
						i = i + 1
						return t[i-1]
					end
					i = i + 1
				until i > n
			end
		else
			return function()
				i = i+1
				if i <= n then
					return t[i]
				end
			end
		end
	end
end

function mtSet.__tolua(set)
	local t = rawget(set,"t")

	return function(s,i)
		tolua(t,i)
	end
end

function mtSet.__tostring(set)
	local t = rawget(set,"t")
	local head,itemhead,itemsep,itemtail,tail

	if type(fmt) == "table" then
		head,itemhead,itemsep,itemtail,tail = fmt.ha, fmt.ih, fmt.is, fmt.it, fmt.ta
	else
		head,itemhead,itemsep,itemtail,tail = "{\n", "\t'", ",\n", "'", "\n}\n"
	end

	if head == nil then head = "{\n" end
	if itemhead == nil then itemhead = "\t'" end
	if itemsep == nil then itemsep = ",\n" end
	if itemtail == nil then itemtail = "'" end
	if tail == nil then tail = "\n}\n" end

	return function()
		local s = head
		for k,v in pairs(t) do
			s = s .. itemhead .. tostring(v) .. itemtail .. itemsep
		end
		s = string.gsub(s, itemsep .. "$","")
		return s .. tail
	end

end


function streq(l,r)
	return tostring(l) == tostring(r)
end

function strlt(l,r) 
	return tostring(l) < tostring(r)
end

function mtSet.__eq() return streq end
function mtSet.__lt() return strlt end

function Set(t,init, ha, ih, is, it, ta)
	local set = {t={},h={}, fmt = {ha = ha, ih = ih, is = is, it = it, ta = ta}}

	for k,f in pairs(mtSet) do
		set[k] = f(set)	
	end

	set.__tostring = mtSet.__tostring(set,set.fmt)

	if type(t) == "table" then
		local insert = mtSet.insert(set,"Ssss")

		for	i,v in pairs(t) do
			if type(init) == "function" then
				insert(init(v))
			else
				insert(v)
			end
		end
	end
	
	setmetatable(set,set)
	return set
end


function each(t,cond)
	local i = 0
	local n = table.getn(t)
	local tc = type(cond)

	if tc == 'nil' then
		return function()
			i = i + 1
			if i <= n then
				return rawget(t,i)
			end
		end
	elseif tc == 'function' then
		return function()
			while i <= n do
				i = i + 1
				local o = rawget(t,i)

				if cond(o) then
					return o
				end
			end
		end
	elseif tc == "string" then
		return function()
			while i <= n do
				i = i + 1
				local o = rawget(t,i)

				if string.match(tostring(o),cond) then
					return o
				end
			end
		end
	else
		error("junk as each cond")
	end
end

mtSymStr.__call = each

function mtSymStr.__tostring(symstr)
	local s = ""
	for sym in each(symstr) do
		s = s .. sym .. " "
	end
	
	return string.gsub(s," $","")
end

function SymStr(...)
	if type(arg[1]) == "table" then
		return setmetatable(arg[1],mtSymStr)
	else
		arg.n = nil
		return setmetatable(arg,mtSymStr)
	end
end


function mtLr0.__tostring(lr0)
	local s = lr0[1][1] .. " ::= " 
	local dp = lr0[2]
	local len = table.getn(lr0[1][2])
	
	for i=1, len do
		if dp == i - 1 then
			s = s .. " ."
		end
		s = s .. " " .. lr0[1][2][i]
	end
	
	if dp == len then
		s = s .. " ."
	end
	
	return s
end

function mtLr0.__call(o,n)
	if n == nil then
		n = 1
	end
	return o[1][2][o[2]+n]
end

mtLr0.__lt = strlt
mtLr0.__eq = streq

function mtLr0.dup(o,j)
	if j == nil then
		j = 0
	end
	local lr = {o[1],o[2]+j}
	setmetatable(lr,mtLr0)
	return lr
end

mtLr1.__lt = mtLr0.__lt
mtLr1.__eq = mtLr0.__eq

function mtLr1.dup(o,j)
	if j == nil then
		j = 0
	end
	
	return setmetatable({o[1],o[2]+j,o[3]},mtLr1)
end

function mtLr1.__call(o,n)
	if n == nil then
		n = 1
	end

	return o[1][2][o[2]+n]
end

function mtLr1.__tostring(lr1)
	local s = lr1[1][1] .. " ::= " 
	local dp = lr1[2]
	local len = table.getn(lr1[1][2])

	for i=1, len do
		if dp == i - 1 then
			s = s .. " ."
		end
		s = s .. " " .. lr1[1][2][i]
	end

	if dp == len then
		s = s .. " ."
	end

	return s .. ", " .. lr1[3]
end

function mtLr1.beta(lr1,n,...)
	local s = {}

	if n == nil then
		n = 0
	end

	for i = lr1[2]+n+1,table.maxn(lr1[1][2]) do
		table.insert(s,lr1[1][2][i])
	end

	if arg.n > 0 then
		for i=1,arg.n do
			table.insert(s,arg[i])
		end
	end

	return setmetatable(s,mtSymStr)
end

function mtRule.lr1(rule,idx,term)
	if idx == nil then idx = 0 end

	if idx > table.getn(rule[2]) + 1 then
		error("lr1 idx higher than number of symbols in string")
	end

	return setmetatable({rule,idx,term},mtLr1)
end

function mtRule.lr0(rule,idx)
	if idx == nil then idx = 0 end
	if idx > table.getn(rule[2]) + 1 then
		error("lr0 idx higher than number of symbols in string")
	end
	
	local t = {rule,idx}
	setmetatable(t,mtLr0)
	return t
end

function mtRule.beta(rule,sym)
	local t = rule[2]
	local n = table.getn(t)
	local p = 1

	return function()
		local add_to = nil

		for i=p,n do
			if type(add_to) == "table" then
				table.insert(add_to, t[i])
			else
				if t[i] == sym then
					p = i+1
					add_to = {}
				end
			end
		end
		
		if add_to ~= nil then
			setmetatable(add_to,mtSymStr)
		end

		return add_to
	end
end


function mtRule.lr0s(rule)
	local s = Set({},'','','\n','','\n')
	for r=0,table.getn(rule[2]) do
		s.insert(rule:lr0(r))
	end
	return s
end

function mtRule.__tostring(rule)
	return rule[1] .. " ::= " .. tostring(rule[2])
end

function Rule(sym,symstr,orig_idx,prec)
	local t = {sym,SymStr(symstr),orig_idx,prec}
	setmetatable(t,mtRule)
	return t
end


function mtGrammar.__tostring(g)
	local s = ""
	s = s .. "rules:\n"
	
	if g.rules == nil then
		g:make_base()
	end

	for i,r in pairs(g.rules) do
		s = s .. "(" .. i ..")\t" .. tostring(r) .. "\n"
	end

	s = s .. "\nall items:\n" .. tostring(g.all_items)

	s = s .. "all symbols:\n"
	for a,n in pairs(g.sym) do
		s = s .. "\t" .. a .. "\t" .. n .. "\n"
	end

	s = s .. "terminals:\n"
	for a,n in pairs(g.term) do
		s = s .. "\t" .. a .. "\t" .. n .. "\n"
	end

	s = s .. "non terminals:\n"
	for a,n in pairs(g.nterm) do
		s = s .. "\t" .. a .. "\t" .. n .. "\n"
	end

	s = s .. "Firsts:\n"
	for a,n in pairs(g.firsts) do
		s = s .. "\t" .. a .. "\t" .. tostring(n) .. "\n"
	end

	s = s .. "Follows:\n"
	for a,n in pairs(g.follows) do
		s = s .. "\t" .. a .. "\t" .. tostring(n) .. "\n"
	end

	return s
end

function mtGrammar.slr_closure(grammar,I)
-- page 223
-- Fig. 4.33
	local J = Set({},nil,"{\n","   ","","\n","}\n")
	J.insert_set(I)

	repeat
		for lr0_a in J() do
			local target = lr0_a()
			for r in each(grammar.rules) do
				if r[1] == target then
					J.insert(r:lr0(0))  --<<<< ??? 0 from where
				end
			end
		end
	until not J.has_changed()
	
	return J
end


function mtGrammar.slr_goto(gr,I,X)
-- page 224
	local J = Set()
	local If = Set()

	local function condition(item)
		local c1 =  "([%w#]) ::= .-" .. X .. " %." -- A -> a B . b 

		local a = string.match(tostring(item),c1)

		if a == nil then return false end

		local c2 = a .. " ::= .* %. " .. X -- A -> a . B b 

		for i in I(c2) do
			return true
		end
	end
	
	for i in I() do
		If.insert(i:dup(1))
	end

	for j in If(condition) do J.insert(j) end
	
	return gr:slr_closure(J)
end

function mtGrammar.make_slr_items(gr)
-- page 224
-- Fig 4.34
	local C = Set({gr:slr_closure(Set({gr.rules[1]:lr0(0)}))})

	repeat
		for I in C() do
			for X,Xi in pairs(gr.sym) do
				local g = gr:slr_goto(I,X)

				if g.len() > 0 then
					C.append(g)
				end
			end
		end

	until not C.has_changed()

	local Ct = {}
	for i in C() do table.insert(Ct,i) end

	gr.slr_canonical = Ct
end

function mtGrammar.make_firsts(gr)
	local firsts = {}

	for sym,i in pairs(gr.sym) do
		   firsts[sym] = gr:first(SymStr({sym}))
	end

	gr.firsts = firsts
end

function mtGrammar.first(gr,symstr,stk)
-- page 189 
	local s = Set({},nil,"",""," ","","")

	if stk == nil then
--		print(symstr)
		stk = {}
	end

	local function stacked(s) 
		for i in each(stk) do
			if i == s then return true end
		end
	end
			
	table.insert(stk,symstr[1])

	if #symstr ==  0 then
		s.insert('&')
		return s
	end

	if #symstr == 1 and gr.term[symstr[1]] ~= nil then
		s.insert(symstr[1])
		return s
	end

	for rule in each(gr.rules) do
		if rule[1] == symstr[1] and not stacked(rule[2][1]) then
			s.insert_set(gr:first(SymStr({rule[2][1]}),stk))
		end
	end
	
	return s
end

function mtGrammar.make_follows(gr)
-- page 189
	local follows = {}

	for A,i in pairs(gr.sym) do
		follows[A] = Set({},nil,"",""," ","","")
	end
	
	follows[gr.rules[1][1]].insert('@')

	function some_has_changed()
		local has_changed = false

		for i,j in pairs(follows) do
			if j.has_changed() then
				has_changed = true
			end
		end
		return has_changed
	end

	repeat
		for A,i in pairs(gr.sym) do
			for B,i in pairs(gr.sym) do

				for rule in each(gr.rules,A .. " ::= .*" .. B .. " .*") do
					for str in rule:beta(B) do
						if str[1] ~= nil then
							local first_beta = gr.firsts[str[1]].copy()
							first_beta.remove('&')
							follows[B].insert_set(first_beta)
						end
					end
				end

				function A_aB(r) if string.match( tostring(r), A .. " ::= .*" .. B) then return true end end
				for rule in each(gr.rules,A_aB) do
					for str in rule:beta(B) do
						if str[1] ~= nil then
							local first_beta = gr.firsts[str[1]].copy()
							if first_beta.contains('&') then
								follows[B].insert_set(follows[A])
							end
						else
							follows[B].insert_set(follows[A])
						end
					end
				end
			end
		end

	until not some_has_changed()

	gr.follows = follows
end

function mtGrammar.slr_getIj(gr,I,a)
	for i,J in pairs(gr.slr_canonical) do
		if gr:slr_goto(I,a) == J then
			return i
		end
	end
end


function mtGrammar.make_slr_table(gr)
-- Page  227-228
-- Alg 4.8 (2)
	gr:make_base()
	gr:make_slr_items()

	local state_tbl = {}
	gr.state_table = state_tbl
	gr.initial_state  = 1


	local function get_reduce(lr0)
		local R = lr0[1]
		for i,r in pairs(gr.rules) do
			if r == R then
				return -i
			end
		end

		error("no such rule: " .. tostring(R))
	end

	local function insert_action(i,a,j)
		local sa = state_tbl[i].action

		if sa[a] == nil then
			sa[a] = j
		else
			gr:resolve_conflict(i,a,j)
		end
	end

	for i, I in pairs(gr.slr_canonical) do
		state_tbl[i] = {goto = {}, action = {}}

		for a,ai in pairs(gr.term) do
			for lr0 in I("::=.* %. " .. a) do
				local j = gr:slr_getIj(I,a)
				if j ~= nil then
					insert_action(i,a,j)
				end
			end
		end

		for lr0 in I("::= .* %.$") do
			local A = lr0[1][1]
			if A ~= '#' then
				for a in gr.follows[A]() do
					insert_action(i,a,get_reduce(lr0))
				end
			end
		end

		local accept_is_in_i = false

		for lr0 in I("#.*%.$") do
			accept_is_in_i = true
		end

		if accept_is_in_i then
			insert_action(i,'@',0)
		end
	end

	for i, I in pairs(gr.slr_canonical) do
		for a,ai in pairs(gr.nterm) do
			local j = gr:slr_getIj(I,a)
			if j ~= nil then
				if state_tbl[i].goto[a] == nil then
					state_tbl[i].goto[a] = j
				else
					error("goto["..i..','..a..'] exists already, is something wrong with the grammar?')
				end
			end
		end
	end

	for i, I in pairs(gr.slr_canonical) do
		for lr0 in I("# ::= %. " .. gr.start_sym .. "$") do
			gr.initial_state = i
		end
	end
	
end



function mtGrammar.can_closure(gr,I)
	J = I.copy()

	repeat
		for i in J( " ::= .-%. .*, " ) do
			local A = i[1][1]
			local B = i()
			local a = i[3]
			local beta = i:beta(1,a)
			for r in each(gr.rules, B .. " ::=" ) do
				local firsts = gr:first(beta)

				for b in firsts() do
					local lr1 = r:lr1(0,b)

					if not J.contains(lr1) then
						J.insert(lr1)
					end
				end
			end
		end
	until not J.has_changed()

	return J
end

function mtGrammar.can_goto(gr,I,X)
	local J = Set()
	for i in I("::=(.*) %. " .. X) do
		local j = i:dup(1)
		if j ~= nil then
			J.insert(j)
		end
	end

	return gr:can_closure(J)
end

function mtGrammar.can_items(g)
	local I0 = g:can_closure(Set({g.rules[1]:lr1(0,'@')}))
	local C = Set({I0})
	local Ct = {I0}

	repeat
		for I in C() do
			for X,x in pairs(g.sym) do
				local goto = g:can_goto(I,X)
				if goto.len() > 0 then
					if C.insert(goto) then
						table.insert(Ct,goto)
					end
				end
			end
		end
	until not C.has_changed()
	return Ct
end

function mtGrammar.make_can_table(gr)
   gr:make_base()

	local can_table = {}
	local C = gr:can_items()

	gr.state_table = can_table
	gr.initial_state  = 1

	local function is_goto_j(I,a)
		for j,g in pairs(C) do
			if g == gr:can_goto(I,a) then
				return j
			end
		end
	end

	local function insert_action(i,a,j)

		if can_table[i] == nil then
			can_table[i] = {action={}, goto={}}
		end

		if can_table[i].action[a] == nil then
			can_table[i].action[a] = j
		else
			gr:resolve_conflict(i,a,j)
		end
	end

	local function get_reduce(r1)
		for i,r2 in pairs(gr.rules) do
			if r1 == r2 then
				return -i
			end
		end
	end

	for i,c in pairs(C) do

		if can_table[i] == nil then
			can_table[i] = {action={}, goto={}}
		end

		for lr1 in c() do
			local A = lr1[1][1]
			local a = lr1()
			local beta = lr1:beta(1)
			local b = lr1[3]

			if gr.term[a] ~= nil then
				local j = is_goto_j(c,a)
				if j ~= nil then 
					insert_action(i,a,j)
				end
			elseif a == nil and A ~= '#' then
				insert_action(i,b,get_reduce(lr1[1]))
			elseif a == nil and A == '#' then
				insert_action(i,'@',0)
			elseif A == '#' then 
				gr.initial_state = i
			end
		end

		for A,id in pairs(gr.nterm) do
			local j = is_goto_j(c,A)
			if j ~= nil then
				can_table[i].goto[A] = j
			end
		end
	end
end


function mtGrammar.resolve_conflict(gr,i,tok,a1)
	local a2 = gr.state_table[i].action[tok]
	local c = {'conflict',false,i,tok,{a1,a2}}
	table.insert(gr.log,c)

	if a1 < 0 and a2 < 0 then --  reduce/reduce conflict:
	--  rule listed first is chosen
		if (a1 <= a2) then
			gr.state_table[i].action[tok] = a1
		else
			gr.state_table[i].action[tok] = a2
		end
		
		local res

		if gr.state_table[i].action[tok] > 0 then
			res = "shift(" .. gr.state_table[i].action[tok] .. ")"
		else
			res = "reduce(" .. -gr.state_table[i].action[tok] .. ")"
		end

		table.insert(c,"using rule given first into " .. res)
	else  --  shift/reduce conflict:
		local shift_a
		local reduce_a

		if a1 > a2 then
			shift_a = a1
			reduce_a = a2
		elseif a1 < a2 then
			shift_a = a2
			reduce_a = a1
		else
			table.insert(c,"no conflict realy")
			return
		end

		local rule = -reduce_a
	if gr.prec[tok] > gr.prec[rule] then
			gr.state_table[i].action[tok] = shift_a
			c[2] = true
			table.insert(c,"token's higher precedence into shift(" .. shift_a .. ")")
		elseif gr.prec[tok] < gr.prec[rule] then
			gr.state_table[i].action[tok] = reduce_a
			c[2] = true
			table.insert(c,"rule's higher precedence into reduce(" .. rule .. ")")
		else
			if gr.assoc[tok] == 'l' then 
				gr.state_table[i].action[tok] = reduce_a
				c[2] = true
				table.insert(c,"token's left associativity into reduce(" .. rule .. ")")
			elseif gr.assoc[tok] == 'r' then 
				gr.state_table[i].action[tok] = shift_a
				c[2] = true
				table.insert(c,"token's right associativity into shift(" .. shift_a .. ")")
			else
				gr.state_table[i].action[tok] = nil
				c[2] = true
				table.insert(c,"token's non associativity into error")
			end
		end
	end
end

function mtGrammar.display_log(gr)
	local s = ""
	for i,c in pairs(gr.log) do
		if c[1] == 'conflict' then
			s = s .. "conflict in state " .. c[3] .. ", for token " .. c[4] .. " between " 
		
			if c[5][1] > 0 then
				s = s .. "shift(" .. c[5][1] .. ") and "
			else
				s = s .. "reduce(" .. -c[5][1] .. ") and "
			end

			if c[5][2] > 0 then
				s = s .. "shift(" .. c[5][2] .. ") "
			else
				s = s .. "reduce(" .. -c[5][2] .. ") "
			end

			s = s .. "resolved by " ..  c[6] 
		else
			s = s .. "unknown log record: "
			for ij,j in pairs (c) do
				s = s .. tostring(j) .. " "
			end
		end

		s = s .. "\n"
	end

	return s
end

function mtGrammar.lr_table_tostring(gr)
	local s1 =  "grammar:\n"

	for i,r in pairs(gr.rules) do
		s1 = s1 .. "(" .. i .. ")   " .. tostring(r) .. "\n"
	end

	local s =  "state\t"
	local s0 = "     \taction"

	for sym,symid in pairs(gr.term) do
		s = s .. sym .. "\t"
		s0 = s0 .. "\t"
	end

	s = s .. "|\t"
	s0 = s0 .. "|goto"

	for sym,symid in pairs(gr.nterm) do
		s = s .. sym .. "\t"
	end

	s = s1 .. "\n\n" .. s0 .. "\n" .. s .. "\n"

	for stateid,state in pairs(gr.state_table) do
		
		s = s ..  stateid .. '\t'

		for sym,symid in pairs(gr.term) do
			local it = state.action[sym]

			if it == nil then
				it = ""
			elseif it  < 0 then
				it = "r" .. -it
			elseif it == 0 then
						it = "acc"
			elseif it > 0 then
				it = "s" .. it
			end

			s = s .. it .. "\t"
		end
		
		s = s .. "|\t"

		for sym,symid in pairs(gr.nterm) do
			local it = state.goto[sym]

			if it == nil then
				it = ""
			end

			s = s .. it .. "\t"
		end

		s = s .. "\n"
	end

	return s
end

function mtGrammar.__tolua()

	local s = header
	
	s = s .. "\n\n-- Prec:\n"
	for i,p in pairs(prec) do
		s = s .. "  --   " .. p[1] .. ":"
	
		for i,o in pairs(p[2]) do
			s = s .. " " .. o
		end
		s = s .. "\n"
	end
	
	local sua = "\n\n-- User Actions:\n" .. prefix .. "_user_actions = {}\n\n"
	s =  s .. "\n\n-- Grammar:\n" .. prefix .. "_grammar = {\n"
	for i,r in pairs(rules) do
		sua = sua .. "-- (" .. i .. ")  " .. r[1] .. " ::= "
		s =  s .. "   [" .. i .. "] =  { '" .. r[1] .. "', { "
		for i2,sym in pairs(r[2]) do
			sua = sua .. " " .. sym
			s = s .. "'" .. sym .. "',"
		end
		s = s .. "}},\n"
		sua = sua .. "\n" .. prefix .. "_user_actions[" .. i .. "] = function(...)" .. r[3] .. "end\n\n"
	end
	
	s = s .. "}\n" .. sua
	
	return s
end


function mtGrammar.make_symbols(gr)
	gr.nterm = { ["#"] = 1 }
	gr.term = { ["@"] = 2 }
	gr.sym = { ["#"] = 1, ["@"] = 2 }

	local symn = 1
	local ntermn = 1

	for r in each(gr.original_rules) do
		if not string.match(r[1],"^[%w%d_]+$") then error("symbols must be all letters/numbers") end
		
		if gr.nterm[r[1]] == nil then
			if not string.match(r[1],"^[%l_]") then error("non terminals must begin with lowercase or '_'") end
			ntermn = ntermn + 1
			gr.nterm[r[1]] = ntermn
		end

		if gr.sym[r[1]] == nil then
			symn = symn + 1
			gr.sym[r[1]] = symn
		end

		for s in each(r[2]) do
			if gr.sym[s] == nil then
				symn = symn + 1
				gr.sym[s] = symn
			end
		end
	end
	
	local termn = 0
	for s,n in pairs(gr.sym) do
		if gr.nterm[s] == nil and gr.term[s] == nil then
			if not string.match(s,"^[%u_]") then error("terminals must begin with uppercase or '_': " .. s) end
			termn = termn + 1
			gr.term[s] = termn
		end
	end

	setmetatable(gr,mtGrammar)
end

function mtGrammar.make_parser(gr)
	pcall(loadstring(gr:tolua(), "parser_driver"))
end

function mtGrammar.make_debug_parser(gr)
	pcall(loadstring(gr:tolua(nil,true), "parser_driver"))
end

function mtGrammar.__tolua(gr,i)
	return gr:tolua(gr.name)
end

function mtGrammar.tolua(gr,name,debug_parser)
	local s

	if name ~= nil then
		s = gr.header .. "\n"
	else
		s = ''
	end

	if name ~= nil then
		s = s .. "function " .. name .. "()\n\t\t local g = "
		.. tolua(
				 {
					 user_actions = gr.user_actions,
					 rules = gr.rules,
					 state_table = gr.state_table,
					 syntax_error = gr.syntax_error,
					 initial_state = gr.initial_state
				 },
				 "\t\t\t\t") .. "\n\n\t\t--debug print('starting ".. name .." in state: 1')\n"
	else
		s = s .. "function gr.parser(g)\n\t --debug print('starting parser in state: 1')\n"
	end

	s = s .. [[
		
		local maxn = table.maxn
		local remove = table.remove
		local insert = table.insert
		
		local rules = g.rules			
		local state_table = g.state_table
		local user_actions = g.user_actions
		local syntax_error = g.syntax_error
		local stack = { state_table[ g.initial_state ] }
	
		return function (tok,v)
			--debug print( "got token:",tok,v)
		
			if tok == nil then
				tok = "@"
				v = nil
			end
		
			while true do
				--debug local ss = ''
				--debug for i,j in pairs(stack) do if i % 2 == 0 then ss = ss .. j[1] .. " " end end
				--debug print("stack:",ss)
	
				local s = stack[maxn(stack)]
				local action = s.action[tok]

				if action == nil then
					local expected = ""
					for t, n in pairs(s.action) do expected = expected .. t .. " " end
		
					--debug print( "syntax error: expecting for: " .. expected )
					if syntax_error ~= nil then
						syntax_error()
						return
					else
						error("syntax error on '" .. tok .. "' while expecting for " .. expected)
					end
				elseif action == 0 then
					--debug print("accept",stack[maxn(stack)-1][2])
					return stack[maxn(stack)-1][2]
				elseif action > 0 then
					insert(stack,{tok,v})
					insert(stack,state_table[action])
					--debug print("shift: going into state: ", action)
					return
				else
					action = -action
					local rule = rules[action]
					local rule_name = rule[1]
					local rule_action = user_actions[action]
		
					--debug local deb = "reduce: " .. rule_name .. " ==> "

					local rule_len = table.getn(rule[2])
					local vs = {}

					for i=1,2*rule_len do
						local t = remove(stack)
						if i % 2 == 0 then
							--debug deb = deb .. rule[2][i/2] .. " "
							insert(vs,1,t[2])
						end
					end
		
					if type(rule_action) == "function" then
						val = rule_action(vs[1],vs[2],vs[3],vs[4],vs[5],vs[6],vs[7],vs[8],vs[9],vs[10])
					end

					s = stack[maxn(stack)]
					insert(stack,{rule_name,val})
					insert(stack,state_table[ s.goto[rule_name] ])
		
					--debug print(deb .. ". going into state: " .. s.goto[rule_name] )
				end
			end
		end
	end
]]  -- '

	if debug_parser then
		s = string.gsub(s,"%-%-debug ",'')
	end
	
	if name ~= nil then
		s = s .. gr.footer .. "\n"
	end

	return s
end

mtUA = {}
function mtUA.__tolua(ua)
	local s = '\n-- user actions \n{\n'
	for i,code in pairs(ua) do
		s = s .. "[" .. i .. "] = " ..  code .. "\n,\n"
	end
	return s .. "\n}\n"
end

					
function mtGrammar.make_rules(gr)
	gr.all_items = Set()
	gr.rules = { Rule("#",{gr.original_rules[1][1]}) }
	gr.user_actions = { gr.accept[1] }
	gr.start_sym = gr.original_rules[1][1]


	gr.all_items.insert(lr0)

	for i,rs in pairs(gr.original_rules) do
		local r = Rule(rs[1],rs[2],i,rs[3])
		table.insert(gr.user_actions,rs[4])
		table.insert(gr.rules,r)

		for lr0 in r:lr0s()() do
			gr.all_items.insert(lr0)
		end
	end
	setmetatable(gr.user_actions,mtUA)

end

function mtGrammar.make_precedence(gr)
	gr.prec = {}
	gr.assoc = {}
	
	for k,v in pairs(gr.original_precedence) do
		for k2,v2 in pairs(v[2]) do
			gr.assoc[v2] = v[1]
			gr.prec[v2] = k
		end
	end

	for sym,ids in pairs(gr.sym) do
		if gr.prec[sym] == nil then 
			gr.prec[sym] = -1
		end
	end

	for ruleid,rule in pairs(gr.rules) do
		local rightmost_sym = nil
		
		if rule[4] ~= 0 then
			rightmost_sym = rule[4]
		else
			for sym in each(rule[2]) do
				if gr.term[sym] ~= nil then
					rightmost_sym = sym
				end
			end
		end

		if rightmost_sym == nil then
			rightmost_sym = '&'
		end

		if gr.prec[rightmost_sym] ~= nil then
			gr.prec[ruleid] = gr.prec[rightmost_sym]
		else
			gr.prec[ruleid] = -1
		end
	end
		
end

function mtGrammar.make_base(gr)
	gr.log = {}

	gr:make_rules()
	gr:make_symbols()
	gr:make_precedence(prec)
	gr:make_firsts()
	gr:make_follows()
end

function mtGrammar.compile_slr(gr)
	gr:make_base()
	gr:make_slr_items()
	gr:make_slr_table()
	gr:make_parser()
end

function mtGrammar.compile_canonical(gr)
	gr:make_base()
	gr:make_caonical_items()
	gr:make_canonical_table()
	gr:make_parser()
end

function mtGrammar.save_parser(gr,parsername,filename,debugging)
	local fh = io.open(filename,'w')

	if header == nil then
		header = "-- this code is automatically generated by LuaYacc, DO NOT MODIFY\n"
	end

	fh:write( gr:tolua(parsername,debugging) )
	fh:close()
end
		
		
mtFN = {}
function mtFN.__tolua(fn)
	return fn[1]
end

function Grammar(g,p,h,f,a,s)
	local gr = {}
	gr.original_rules = g

	if p ~= nil then
		gr.original_precedence = p
    else
		gr.original_precedence = {}
	end

	if h == nil then
		h = "" 
	end

	gr.header = h
	if f == nil then
		f = "" 
	end

	gr.footer = f

	if a == nil then
		a = {"function(v) return v end"}
	else
		a = {a}
	end

	gr.accept = a
	setmetatable(gr.accept,mtFN)
		
	if s == nil then
		s = {"function() error('syntax error') end"}
	else
		s = {s}
	end

	gr.syntax_error = s
	setmetatable(gr.syntax_error,mtFN)

	setmetatable(gr,mtGrammar)

	return gr
end


function LuaYacc(in_filename,out_filename,debuggable)
	local state = 'outside'
	local prec = {}
	local rules = {}
	local tokens = {}
	local patterns = {}
	local ignores = {}

	local body = ''
	local text = ''
	local header = '-- automatically generated by LuaYacc DO NOT MODIFY! \n'
	local rule
	local name = "yyparse"
	local footer = ''
	local accept = 'nil'
	local syntax_error = 'nil'
	local lexical_error = 'nil'

	if out_filename == nil then
		if string.match(in_filename,"%.yl") then
			out_filename = string.gsub(in_filename,"%.yl",".lua")
		else
			error("cannot guess output filename")
		end
	end


	local function add_oplist(assoc,str)
		local lst = {}
		local p = {assoc,lst}

		string.gsub(str,"(%u+)", function(s) table.insert(lst,s) end )
		
		table.insert(prec,p)
	end
	
	local token_id = 0
	local function next_tok_name()
		token_id = token_id + 1
		return "_" .. token_id .. "_"
	end

	local function parse_rhs(rhs)
		local symstr = {}
		local argnames = {}
		local i = 0
		local argname
		local prec = 0
		local orig_rhs = rhs
		
		rhs = string.gsub(rhs,"%%prec%s+([%u_]+)",function (p) prec = p;  return "" end)
		rhs = string.gsub(rhs,"^%s*(.-)%s*$","%1")
		
		function unnamed_symbol(s,n)
			sym = s
			return ""
		end


		function named_symbol(s,n)
			sym = s
			argname = n
			return ""
		end

--		function unnamed_pattern(p)
--			sym = next_tok_name()
--			p = string.gsub(p,"^.(.-).$","%1")
--			patterns[p] = sym
--			table.insert(tokens,{"^(" .. p .. ")",sym})
--			return ""
--		end

--		function named_pattern(p,n)
--			sym = next_tok_name()
--			p = string.gsub(p,"^.(.-).$","%1")
--			patterns[p] = sym
--			table.insert(tokens,{"^(" .. p .. ")",sym})
--			argname = n
--			return ""
--		end

		local actions = {
			{"^%s*(%w+)%s*%(%s*([%l_]+)%)", named_symbol},
--			{"^%s*('[^']+')%s*%(%s*([%l_]+)%)", named_pattern},
--			{'^%s*("[^"]+")%s*%(%s*([%l_]+)%)', named_pattern},
--			{"^%s*(%b<>)%s*%(%s*([%l_]+)%)", named_pattern},
			{"^%s*(%w+)%s*", unnamed_symbol},
--			{"^%s*('[^']+')%s*", unnamed_pattern},
--			{'^%s*("[^"]+")%s*', unnamed_pattern},
--			{"^%s*(%b<>)%s*", unnamed_pattern}
		}

		while #rhs > 0 do
			i = i+1
			argname = "__" .. i
			sym = nil
			
			for j,act in pairs(actions) do
				rhs = string.gsub(rhs,act[1],act[2], 1)
				if sym ~= nil then break end
			end
			
			if sym == nil then
				error("cannot match pattern in rule rhs: '" .. orig_rhs .. "'")
			end

			table.insert(argnames,argname)
			table.insert(symstr,sym)

			if i > 100 then
				error("hanging while parsing rule's rhs : '" .. orig_rhs .. "'")
			end
		end
		
		return symstr, argnames, prec
	end

	local function start_rule_with_body(lhs,rhs)
		local symstr
		local argnames
		local prec

		symstr, argnames, prec = parse_rhs(rhs)
		rule = {lhs,symstr,prec}
		text = "\n -- " .. lhs .. " ::= " .. rhs .. "\n function("
		
		local argstr = ""
		for i, a in pairs(argnames) do
			argstr = argstr .. a .. ", "
		end
		argstr = string.gsub(argstr,", $","")

		text = text .. argstr .. ")\n"

		table.insert(rules,rule)
	end
	
	local function finish_rule()
			text = text .. "\nend\n"

			table.insert(rule,text)
	end



	function rule_without_body(lhs,rhs)
		local symstr
		local argnames
		local prec
		local fntxt = "\n -- " .. lhs .. " ::= " .. rhs .. "\n function("

		symstr, argnames, prec = parse_rhs(rhs)
		rule = {lhs,symstr,prec}

		local has_usernames = false
		local vs = ""
		for i, v in pairs(argnames) do
			vs = vs .. v .. ", "
			if not string.match(v,"^__%d+$") then
				has_usernames = true
			end
		end
		vs = string.gsub(vs,", $","")

		if has_usernames then
			fntxt = fntxt .. vs .. ")\n\treturn {"
			for k, v in pairs(argnames) do
				if not string.match(v,"^__%d+$") then
					fntxt = fntxt .. v .. " = " .. v .. ", "
				end
			end
			fntxt = fntxt .. "}\nend\n"
		else 
			fntxt = fntxt .. "...)\n\targ.n = nil\n\treturn arg\nend\n"
		end

		table.insert(rule,fntxt)
		table.insert(rules,rule)

	end

	function pattern_list(txt)
		local t = {}
		local i = 0
		while #txt > 0 do
			i = i + 1
			for j,pat in pairs({ "^%s*('[^']-')%s*", '^%s*("[^"]-")%s*', "^%s*(%b<>)%s*" }) do
				txt = string.gsub(txt,pat,
								  function(a)
										a = string.gsub(a,"^.(.*).$","%1");
										table.insert(t,a);
										return ""
								  end)
			end

			if i > 100 then error("hanging") end
		end
		return t
	end

	function ignore(patternstr)
		for i,p in pairs(pattern_list(patternstr)) do
			table.insert(ignores,p)
		end
	end

	function token(sym,patternstr)
		local ps = pattern_list(patternstr)

		for i,p in pairs(ps) do
			patterns[p] = sym
			table.insert(tokens,{"^(" .. p .. ")",sym})
		end

	end


	local fsm = {
		outside = {
			['^%-%-header'] = function() state = 'in_hdr'; header = '' end,
			['^%-%-footer'] = function() state = 'in_ftr'; footer = '' end,
			['^%-%-syntax_error'] = function() state = 'in_syn'; syntax_error = 'function()\n' end,
--			['^%-%-lexical_error'] = function() state = 'in_lex'; lexical_error = 'function()\n' end,
			['^%-%-accept%s+([%l_]+)'] = function(varname) state = 'in_acc'; accept = 'function('..varname..')\n' end,
			['^%-%-name%s+(%w+)'] = function(n) name = n end,
			['^%-%-left%s+(.*)'] = function(oplist_str) add_oplist('l',oplist_str) end,
			['^%-%-right%s+(.*)'] = function(oplist_str) add_oplist('r',oplist_str) end,
			['^%-%-nonassoc%s+(.*)'] = function(oplist_str) add_oplist('n',oplist_str) end,
			['^%s*([%l_]+)%s*::=(.-)%-%-'] = rule_without_body ,
			['^%s*([%l_]+)%s*::=(.*)'] = function(lhs,rhs) state = 'in_fn'; start_rule_with_body(lhs,rhs) end,
--			['^%-%-ignore%s+(.*)'] = ignore,
--			['^%-%-token%s+(%u*)(.*)'] = token,
		},
		
		in_fn = {
			['^%-%-$'] = function() state = 'outside'; finish_rule() end,
			['^(.*)$'] = function(s) text = text .. s .. "\n" end
		},
	
		in_hdr = {
			['^%-%-$'] = function() state = 'outside'; header = header ..  "\n" end,
			['^(.*)$'] = function(s) header = header .. s .."\n" end
		},
		
		in_ftr = {
			['^%-%-$'] = function() state = 'outside'; footer = footer ..  "\n"  end,
			['^(.*)$'] = function(s) footer = footer .. s .."\n" end
		},
		
		in_syn = {
			['^%-%-$'] = function() state = 'outside'; syntax_error =  syntax_error ..  "\nend\n" end,
			['^(.*)$'] = function(s) syntax_error = syntax_error .. s .."\n" end
		},

		in_lex = {
			['^%-%-$'] = function() state = 'outside'; lexical_error =  lexical_error ..  "\nend\n" end,
			['^(.*)$'] = function(s) lexical_error = lexical_error .. s .."\n" end
		},
		
		in_acc = {
			['^%-%-$'] = function() state = 'outside'; accept =  accept ..  "\nend\n"  end,
			['^(.*)$'] = function(s) accept = accept .. s .."\n" end
		},
		
	}
	
	for line in io.lines(in_filename) do
		-- print(state,tolua(fsm[state]))

		for pattern,action in pairs(fsm[state]) do
			if string.match(line,pattern) ~= nil then
				string.gsub(line,pattern,action,1);
				break
			end
		end
	end
	
	header = header .. "\n"

--print("==================================================header\n" .. header)
--print("==================================================footer\n" .. footer)
--print("==================================================syntax_error\n" .. syntax_error)
--print("==================================================lexical_error\n" .. lexical_error)
--print("==================================================accept\n" .. accept)
--print("==================================================rules\n" .. tolua(rules))
--print("==================================================prec\n" .. tolua(prec))
--print("==================================================tokens\n" .. tolua(tokens))
--print("==================================================patterns\n" .. tolua(patterns))
--print("==================================================ignores\n" .. tolua(ignores))

	local g =  Grammar(rules,prec,header,footer,accept,syntax_error)
	g:make_can_table()
	g:save_parser(name,out_filename,debuggable)

	return g
end

function test()
	g4_3 = {
		{ 'e', {'e', 'P', 'e'} },
		{ 'e', {'e', 'T', 'e'} },
		{ 'e', {'M', 'e'} },
		{ 'e', {'O','e','C' } },
		{ 'f', {'N'} },
	}

	g4_6 = {
		{'s',{'s','O','s','C'}},
		{'s',{}}
	}

	g4_10 = {
		{ 'e', {'e', 'P', 't'} },
		{ 'e', {'t'} },
		{ 't', {'t', 'T', 'f'} },
		{ 't', {'f'} },
		{ 'f', {'O','e','C' } },
		{ 'f', {'N'} },
	}

	g4_11 = {
		{ 'e1', {'t1','e2'}},
		{ 'e2', {'P','t1','e2'}},
		{ 'e2', {}},
		{ 't1', {'f','t2'}},
		{ 't2', {'M','f','t2'}},
		{ 't2', {}},
		{ 'f', {'O','e1','C' }},
		{ 'f', {'N'}},
	}

	g4_12 = {
		{'e', {'e', 'M', 'b'}},
		{'e', {'e', 'p', 'b'}},
		{'e', {'b'}},
		{'b', {'Z'}},
		{'b', {'O'}},
	}

	g4_17 = {
		{ 'e', {'e', 'P', 'e'} },
		{ 'e', {'e', 'M', 'e'} },
		{ 'e', {'e', 'T', 'e'} },
		{ 'e', {'e', 'D', 'e'} },
		{ 'e', {'e', 'E', 'e'} },
		{ 'e', {'O','e','C' } },
		{ 'e', {'M', 'e'} },
		{ 'e', {'N'} },
	}

	g4_22 = {
		{ 'e', {'e', 'P', 'e'} },
		{ 'e', {'e', 'T', 'e'} },
		{ 'e', {'O','e','C' } },
		{ 'e', {'N'} },
	}

	g4_21 = {
		{'s', {'c', 'c'}},
		{'c', {'C','c'}},
		{'c', {'D'}}
	}

	return Grammar(g4_21)
end
